/*------------------------------------------------------------------------------
| Module:      lbpi.c                                                          |
| ---------------------------------------------------------------------------- |
| Purpose:     Little-bitty perl interpreter                                   |
|                                                                              |
|     Runs an embeded perl on the given script.  Only really necessary so      |
|     that shared objects get loaded in the right order (which they don't      |
|     if the perl interpreter is run directly).                                |
|                                                                              |
| Functions:                                                                   |
|                                                                              |
|     main                                                                     |
|                                                                              |
------------------------------------------------------------------------------*/



/*----------------------------------------------------------------------
| Includes                                                             |
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>

#define _GNU_SOURCE
#include <getopt.h>

#include "EXTERN.h"
#include "perl.h"



/*----------------------------------------------------------------------
| Global/statics                                                       |
----------------------------------------------------------------------*/
static PerlInterpreter  *perl;



/*----------------------------------------------------------------------
| Local prototypes                                                     |
----------------------------------------------------------------------*/
void usage (void);



/*----------------------------------------------------------------------
| Magic incantations to make extensions work                           |
----------------------------------------------------------------------*/
static void xs_init (void);

extern void boot_DynaLoader (CV *cv);
extern void boot_wine       (CV *cv);

extern void xs_init()
{
    char *file = __FILE__;

    newXS ("DynaLoader::boot_DynaLoader", boot_DynaLoader, file);
    newXS ("wine::bootstrap",             boot_wine,       file);
}



    /*--------------------------------------------------------------
    | Code                                                         |
    |                                                              |
    |                                                              |
    --------------------------------------------------------------*/




/*----------------------------------------------------------------------
| Function:    main                                                    |
| -------------------------------------------------------------------- |
| Purpose:     Entry point; embed Perl and call perl script            |
|                                                                      |
| Parameters:  argc/argv (see usage() ...)                             |
|                                                                      |
| Returns:     status                                                  |
----------------------------------------------------------------------*/
int main (int argc, char *argv[], char *envp[])
{
    /* Locals */
    char         *script_file    = NULL;
    char         *entry_sub      = strdup ("main");
    struct stat  stat_buf;
    char         *perl_argv[4]   = { "perl", "-x", NULL , NULL };
    int          n_args          = (sizeof (perl_argv) / sizeof (perl_argv[0]) - 1);
    char         *script_argv[1] = { NULL };

/* Begin main */

    /*--------------------------------------------------------------
    | Process args
    --------------------------------------------------------------*/
    while (1)
    {

        /*--------------------------------------------------------------
        | Locals, and long option defn's
        --------------------------------------------------------------*/
        int  c;
        int  option_index = 0;
        static struct option long_options[] =
        {
            { "perl-script", required_argument, NULL, 'p' },
            { "entry-sub",   required_argument, NULL, 'e' },
            { "help",        no_argument,       NULL, 'q' },
            { NULL,          0,                 NULL, 0   }
        };

        /*--------------------------------------------------------------
        | Grab 'em
        --------------------------------------------------------------*/
        if ((c = getopt_long (argc, argv, "p:f:e:q", long_options, &option_index)) == -1)
        {
            break;
        }
        switch (c)
        {
            case 'p':
            case 'f':
                script_file = strdup (optarg);
                break;
            case 'e':
                entry_sub = strdup (optarg);
                break;
            case 'q':
            default:
                usage ();
                goto EXIT;
        }

    }  /* end while */

    /*--------------------------------------------------------------
    | If there are any args left after getopt() processing, then
    | the first one is taken to be the script file
    --------------------------------------------------------------*/
    if (optind < argc)
    {
        script_file = strdup (argv[optind]);
    }

    /*--------------------------------------------------------------
    | If no script file specified, exit
    --------------------------------------------------------------*/
    if (! script_file)
    {
        usage ();
        goto EXIT;
    }

    /*--------------------------------------------------------------
    | Check for script file
    --------------------------------------------------------------*/
    if (stat (script_file, &stat_buf) == -1)
    {
        printf ("Can't stat script file [%s]\n", script_file);
        goto EXIT;
    }

    /*--------------------------------------------------------------
    | Allocate a perl interpreter
    --------------------------------------------------------------*/
    if ((perl = perl_alloc ()) != NULL)
    {
        perl_construct (perl);

        /*--------------------------------------------------------------
        | Parse script file
        --------------------------------------------------------------*/
        perl_argv[n_args - 1] = script_file;
        perl_parse 
        (
            perl,
            xs_init,
            n_args,
            perl_argv,
            envp
        );

        /*--------------------------------------------------------------
        | Call entry sub (no args required)
        --------------------------------------------------------------*/
        perl_call_argv 
        (
            entry_sub,
            G_DISCARD,
            script_argv
        );

        /*--------------------------------------------------------------
        | Clean up
        --------------------------------------------------------------*/
        perl_destruct (perl);
        perl_free (perl);

    }  /* end if (perl interpreter allocated successfully) */


    /*--------------------------------------------------------------
    | Clean up allocations and blow outta here
    --------------------------------------------------------------*/
EXIT:

    if (script_file)
    {
        free (script_file);
    }
    if (entry_sub)
    {
        free (entry_sub);
    }

    exit (0);

} /* End main */




/*----------------------------------------------------------------------
| Function:    usage                                                   |
| -------------------------------------------------------------------- |
| Purpose:     Display usage                                           |
|                                                                      |
| Parameters:  (none)                                                  |
|                                                                      |
| Returns:     (none)                                                  |
----------------------------------------------------------------------*/
void usage (void)
{
    /* Locals */

/* Begin usage */

    printf ("Usage:  lpbi [options] [perl_script]\n");
    printf ("Options:\n");
    printf ("    -p, --perl-script       perl script to run\n");
    printf ("    -e, --entry-sub         entry point subroutine [default: main()]\n");

} /* End usage */
