/*------------------------------------------------------------------------------
| Module:      wine.xs                                                         |
| ---------------------------------------------------------------------------- |
| Purpose:     Perl gateway to wine API calls                                  |
|                                                                              |
| Functions:                                                                   |
|                                                                              |
|     Straight C routines                                                      |
|     -------------------                                                      |
|     wine_init_callback -- wine initialization callback function              |
|                                                                              |
|     XS functions                                                             |
|     ------------                                                             |
|     wine_init     -- call wine initialization function                       |
|     call_wine_API -- call a wine API function                                |
|                                                                              |
------------------------------------------------------------------------------*/



/*----------------------------------------------------------------------
| Includes                                                             |
----------------------------------------------------------------------*/
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"

#include <stdlib.h>
#include <string.h>

#include "wine_perl_gateway.h"



/*----------------------------------------------------------------------
| Other pre-processor stuff                                            |
----------------------------------------------------------------------*/
#define free_ptr(PTR)                                                  \
do                                                                     \
{                                                                      \
    if (PTR)                                                           \
        free (PTR);                                                    \
    (PTR) = NULL;                                                      \
}                                                                      \
while (0);



/*----------------------------------------------------------------------
| Globals                                                              |
----------------------------------------------------------------------*/
char *type_str[] =
{
    "int",
    "ptr"
};




/*----------------------------------------------------------------------
| Function:    wine_init_callback                                      |
| -------------------------------------------------------------------- |
| Purpose:     Callback entry point for PROCESS_InitWinelib()          |
|                                                                      |
| Parameters:  argc, argv                                              |
|                                                                      |
| Returns:                                                             |
----------------------------------------------------------------------*/
int wine_init_callback (int argc, char *argv[])
{
    /* Locals */
    char  *perl_callback;

/* Begin wine_init_callback */

    dSP;

    /*--------------------------------------------------------------
    | Perl callback function should be argv[1]
    --------------------------------------------------------------*/
    perl_callback = argv[1];
    PUSHMARK(SP);
    perl_call_pv (perl_callback, G_DISCARD|G_NOARGS);


    /*--------------------------------------------------------------
    | Ciao
    --------------------------------------------------------------*/
    return (0);

} /* End wine_init_callback */




/*----------------------------------------------------------------------
| XS module                                                            |
|                                                                      |
|                                                                      |
----------------------------------------------------------------------*/
MODULE = wine		PACKAGE = wine		




    # ------------------------------------------------------------------------
    # | Function:    wine_init                                               |
    # | -------------------------------------------------------------------- |
    # | Purpose:     Call C function which calls wine initialization routine |
    # |                                                                      |
    # | Parameters:                                                          |
    # |                                                                      |
    # | Returns:                                                             |
    # ------------------------------------------------------------------------
void
wine_init(perl_callback)
    char  *perl_callback;

    CODE:
    /*--------------------------------------------------------------
    | Begin wine_init
    --------------------------------------------------------------*/

    /* Locals */
    int   argc     = 2;
    char  *argv[3] =
    {
        "perl",
        perl_callback,
        NULL
    };

    /*--------------------------------------------------------------
    | Call init routine
    --------------------------------------------------------------*/
    (void) perl_init_wine (argc, argv);


    /*--------------------------------------------------------------
    | End wine_init
    --------------------------------------------------------------*/




    # ------------------------------------------------------------------------
    # | Function:    call_wine_API                                           |
    # | -------------------------------------------------------------------- |
    # | Purpose:     Call perl_call_wine(), which calls a wine API function  |
    # |                                                                      |
    # | Parameters:  module   -- module (dll) to get function from           |
    # |              function -- API function to call                        |
    # |              ret_type -- return type                                 |
    # |              debug    -- debug flag                                  |
    # |              ...      -- args to pass to API function                |
    # |                                                                      |
    # | Returns:     value returned by API function                          |
    # ------------------------------------------------------------------------
void
call_wine_API(module, function, ret_type, debug, ...)
    char  *module;
    char  *function;
    int   ret_type;
    int   debug;

    CODE:
    /*--------------------------------------------------------------
    | Begin call_wine_API
    --------------------------------------------------------------*/

    /* Local types */
    struct arg
    {
        int           ival;
        char          *pval;
    };

    /* Locals */
    int            n_fixed = 4;
    int            n_args = (items - n_fixed);
    struct arg     *args;
    unsigned long  *f_args;
    int            i, n;
    char           *p;
    SV             *sv;
    unsigned long  r;

    /*--------------------------------------------------------------
    | Allocate some memory for arg processing
    --------------------------------------------------------------*/
    args   = (struct arg *)    calloc (n_args, sizeof (*args));
    f_args = (unsigned long *) calloc (n_args, sizeof (*f_args));

    /*--------------------------------------------------------------
    | Prepare function args
    --------------------------------------------------------------*/
    if (debug)
    {
        printf ("    [wine.xs/call_wine_API()]\n");
    }
    for (i = 0; (i < n_args); i++)
    {
        sv = ST (n_fixed + i);
        if (! SvOK (sv))
            continue;

        /*--------------------------------------------------------------
        | Ref
        --------------------------------------------------------------*/
        if (SvROK (sv))
        {
            sv = SvRV (sv);

            /*--------------------------------------------------------------
            | Integer ref -- pass address of value
            --------------------------------------------------------------*/
            if (SvIOK (sv))
            {
                args[i].ival = SvIV (sv);
                f_args[i] = (unsigned long) &(args[i].ival);
                if (debug)
                {
                    printf ("        [RV->IV] %p\n", f_args[i]);
                }
            }

            /*--------------------------------------------------------------
            | Number ref -- convert and pass address of value
            --------------------------------------------------------------*/
            else if (SvNOK (sv))
            {
                args[i].ival = (unsigned long) SvNV (sv);
                f_args[i] = (unsigned long) &(args[i].ival);
                if (debug)
                {
                    printf ("        [RV->NV] %p\n", f_args[i]);
                }
            }

            /*--------------------------------------------------------------
            | String ref -- pass pointer
            --------------------------------------------------------------*/
            else if (SvPOK (sv))
            {
                f_args[i] = (unsigned long) ((char *) SvPV (sv, PL_na));
                if (debug)
                {
                    printf ("        [RV->PV] %p\n", f_args[i]);
                }
            }
        }

        /*--------------------------------------------------------------
        | Scalar
        --------------------------------------------------------------*/
        else
        {

            /*--------------------------------------------------------------
            | Integer -- pass value
            --------------------------------------------------------------*/
            if (SvIOK (sv))
            {
                f_args[i] = (unsigned long) SvIV (sv);
                if (debug)
                {
                    printf ("        [IV]     %ld (0x%x)\n", f_args[i], f_args[i]);
                }
            }

            /*--------------------------------------------------------------
            | Number -- convert and pass value
            --------------------------------------------------------------*/
            else if (SvNOK (sv))
            {
                f_args[i] = (unsigned long) SvNV (sv);
                if (debug)
                {
                    printf ("        [NV]     %ld (0x%x)\n", f_args[i], f_args[i]);
                }
            }

            /*--------------------------------------------------------------
            | String -- pass pointer to copy
            --------------------------------------------------------------*/
            else if (SvPOK (sv))
            {
                p = SvPV (sv, n);
                if (args[i].pval = calloc (n, 1))
                {
                    memcpy (args[i].pval, p, n);
                    f_args[i] = (unsigned long) args[i].pval;
                    if (debug)
                    {
                        printf ("        [PV]     %p\n", f_args[i]);
                    }
                }
            }
        }

    }  /* end for */

    /*--------------------------------------------------------------
    | Here we go
    --------------------------------------------------------------*/
    r = perl_call_wine
    (
        module,
        function,
        n_args,
        f_args,
        debug
    );

    /*--------------------------------------------------------------
    | Set global error code
    --------------------------------------------------------------*/
    if ((sv = perl_get_sv ("wine::err", FALSE)) != NULL)
    {
        sv_setiv (sv, GetLastError ());
    }

    /*--------------------------------------------------------------
    | Handle modified parameter values
    |
    | There are four possibilities for parameter values:
    |
    |     1) integer value
    |     2) string value
    |     3) ref to integer value
    |     4) ref to string value
    |
    | In cases 1 and 2, the intent is that the values won't be
    | modified, because they're not passed by ref.  So we leave
    | them alone here.
    |
    | In case 4, the address of the actual string buffer has
    | already been passed to the wine API function, which had
    | opportunity to modify it if it wanted to.  So again, we
    | don't have anything to do here.
    |
    | The case we need to handle is case 3.  For integers passed
    | by ref, we created a local containing the initial value,
    | and passed its address to the wine API function, which
    | (potentially) modified it.  Now we have to copy the
    | (potentially) new value back to the Perl variable passed
    | in, using sv_setiv().  (Which will take fewer lines of code
    | to do than it took lines of comment to describe ...)
    --------------------------------------------------------------*/
    for (i = 0; (i < n_args); i++)
    {
        sv = ST (n_fixed + i);
        if (! SvOK (sv))
            continue;
        if (SvROK (sv) && (sv = SvRV (sv)) && SvIOK (sv))
        {
            sv_setiv (sv, args[i].ival);
        }
    }

    /*--------------------------------------------------------------
    | Put appropriate return value on the stack for Perl to pick
    | up
    --------------------------------------------------------------*/
    switch (ret_type)
    {
        case RET_INT:  XSRETURN_IV ((int) r);     break;
        case RET_PTR:  XSRETURN_PV ((char *) r);  break;
    }

    /*--------------------------------------------------------------
    | Free up allocated memory
    --------------------------------------------------------------*/
    for (i = 0; (i < n_args); i++)
    {
        free_ptr (args[i].pval);
    }
    free_ptr (args);
    free_ptr (f_args);


    /*--------------------------------------------------------------
    | End call_wine_API
    --------------------------------------------------------------*/
